/*
 * $RCSfile: msg.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: msg.c,v $
 * Revision 1.78.1.2  1994/06/08  23:18:08  davidm
 * KLUDGE is mach-specific
 *
 * Revision 1.78.1.1  1994/04/25  20:28:55  menze
 * XLeafFun takes a single argument (to match decRefnr)
 *
 * Revision 1.78  1994/04/12  22:39:58  hkaram
 * Removed msgLen (became a macro in msg.h)
 * Removed references to MSGCACHEING
 * Fixed an IRIX compile error in msgnrpop
 *
 * Revision 1.77  1994/04/12  01:07:45  hkaram
 * David's Changes:
 *  Pointers now printed as "%lx" instead of "%x".  The pointer argument
 *  is cast to (u_long) to make this work.
 * My changes:
 * 1. Removed MSG_NEW_ALG ifdefs
 * 2. Removed some dead code
 * 3. Num of nodes in ConstructEmpty = 1 (prev 0)
 * 4. Changed the msgnrpush/pop to be macros for speed
 * 5. Removed TU_UNFOLD #ifdef'ed code. Removed tu_leaf. Now uses the
 *    loop function unconditionally.
 * 6. Changed the msgnrpush/pop to be macros for speed
 * 7. Rearranged the switch statements so that leafs are found
 *    first for the foreach operations.
 * 8. Added fast case in ForEach for a single node msg
 * 9. Stack ownership fix in msgAppend
 * 10. Fixed a bug in msgAppend where the lastStackTailPtr was not set
 *    properly causing consecutive msgAppend's to fail.
 * 11. Fixed a bug in msgPop where a peak was not being handled
 *    correctly after an underflow. Simplified msgPop.
 *    Also laststack ownership is not lost automatically after a msgPop.
 * 12. Fixed msgAssign and msgConstructCopy to copy data structure
 *    instead of item at a time.
 * 13. Fixed the lastStack ownership to be conditionally set in msgJoin
 *
 * Revision 1.76  1994/03/12  22:02:38  davidm
 * Pointers now printed with "%lx" format and cast to (u_long) to make
 * it compile cleanly on Alphas.
 *
 * Revision 1.75  1994/02/05  00:08:33  menze
 *   [ 1994/01/30          menze ]
 *   assert.h renamed xk_assert.h
 *
 * Revision 1.74  1993/12/14  18:27:42  menze
 *   [ 1993/12/14          menze ]
 *   Fixed a '#else' comment
 *
 *   [ 1993/12/13          menze ]
 *   Modifications from UMass:
 *
 *     [ 93/06/19          yates ]
 *     fixed FINDAFREEONE macro
 *
 *   [ 1993/12/11          menze ]
 *   fixed #endif comments
 *
 *   [ 1993/12/01             ho ]
 *   Added cast of silly tu_leaf function in MSG_NON_REC mode.
 *
 *   [ 1993/10/02          menze ]
 *   Added more strict prototypes
 *
 *   [ 1993/09/28          menze ]
 *   Added a cast to a msgForEach call (fix from cs.umass.edu)
 *
 */

/*
 *  The msg tool package supports tree structured message management.
 *
 *  The algorithms were developed by Peter Druschel in C++
 *
 */


#include "xk_assert.h"
#include "xk_debug.h"
#include "platform.h"
#include "x_util.h"
#include "x_libc.h"

#include "msg.h"
#include "msg_internal.h"

/* this compensates for a bug on the Mach platform */
#ifdef MACH3
# define KLUDGE 4
#else
# define KLUDGE 0
#endif

static int	  mpr_used;
static int	  mpr_hiwat;

static int	  mpg_used;
static int	  mpg_hiwat;

static int	  mup_used;
static int	  mup_hiwat;

static int	  mbf_used;
static int	  mbf_hiwat;


static MNodeLeaf   dummyStack;

/* limit for number of nodes allocated to a Msg */
#ifdef CONTROL_RESOURCES
static int numNodesHardLimit = 1000;
#endif /* CONTROL_RESOURCES */
static int numNodesSoftLimit = 100;
static int msgSpace_used, msgSpace_hiwat;




#define ROUND4(len)  ((len + 3) & ~3)

int	tracemsg = 0;

#if defined(__STDC__) || defined(__GNUC__)

static unsigned	atomicDec( unsigned long * );
static unsigned	atomicInc( unsigned long * );
static bool	frag2Buf( char *, long, void * );
static void	incRef( MNode * );
static void	makeContiguous( Msg * );
static void 	msgInternalShow( MNode *, long, long, long);
static bool 	msgInternalForEach( MNode *, long, long, XCharFun, VOID *);
static char *	msgTopUnderflow( Msg *, long );
static MNodePage *newMNodePage( int );
static MNodePair *newMNodePair(long, long, MNode *, long, long, MNode * );
static MNodePage *newMNodeUsrPage(MsgCIPFreeFunc, char *, long );
static char *	msgPushOverflow( Msg *, long );
static MNodePage *newSmallMNodePage( int );
static bool	s_copy( char *, long, void * );


#ifdef MSG_NR_TEST
bool msgnrUserFun(MNode *this, long off, long len, VOID *arg);
void msgnrtest(void);
#endif /* MSG_NR_TEST */

#ifndef MSG_NON_REC
static void	decRef( MNode * );
#endif /* MSG_NON_REC */

#else

static void	makeContiguous();
static void msgInternalShow( );
static bool msgInternalForEach( );

#ifdef MSG_NR_TEST
bool msgnrUserFun();
void msgnrtest();
#endif /* MSG_NR_TEST */

#endif

#ifdef __STDC__
#define __CONCAT(a,b) a##b
#else
#define __CONCAT(a,b) a/**/b
#endif


#ifdef MSG_STATISTICS
#  define RECORD_USAGE(prefix) { 		\
  __CONCAT(prefix,_used)++; 				\
  if ( __CONCAT(prefix,_used) > __CONCAT(prefix,_hiwat) ) {	\
      __CONCAT(prefix,_hiwat) = __CONCAT(prefix,_used);	\
  }						\
  msgSpace_used += __CONCAT(prefix,size); 		\
  if ( msgSpace_used > msgSpace_hiwat ) {	\
      msgSpace_hiwat = msgSpace_used;		\
  }						\
}

#  define RECORD_RELEASE(prefix) { 		\
  __CONCAT(prefix,_used)--; 				\
  msgSpace_used -= __CONCAT(prefix,size); 		\
}						\

#else /* ! MSG_STATISTICS */

#define RECORD_USAGE(prefix)
#define RECORD_RELEASE(prefix)

#endif /* MSG_STATISTICS */


/*
 * Forward:
 */

#ifdef MSG_NON_REC

#define  msgnrpush(stack, item) \
           *stack->top++ = (item); \
           if (stack->top >= stack->limit) msgnrpush_overflow(stack);

#define  msgnrpop(stack, item) \
  	   ( (--(stack->top) >= stack->bottom) ? \
		((*((MNode ** )item) = *(stack->top)) , 1) : 0);

static struct mstack msgStack;

#if defined(__STDC__) || defined(__GNUC__)
static void msgnrstackinit(struct mstack *);
/* 
 * types of user functions called in ForEach()
 */
typedef bool (*XNodeFun)( struct mstack *, MNodePair *, void * );
typedef bool (*XLeafFun)( MNode * );

#else
static void msgnrstackinit();
typedef bool (*XNodeFun)();
typedef bool (*XLeafFun)();
#endif

#endif /* MSG_NON_REC */

/* Msg operations */

/* initializer; must be called before any Msg instances are created 	*/
/* returns 1 after successful intialization				*/

bool msgInit()
{
  xTrace0(msg, TR_FULL_TRACE, "msgInit");

  /* init the dummy stack for empty messages */
  dummyStack.type = t_MNodeDummy;
  dummyStack.refCnt = 1;
  dummyStack.b.leaf.size = 0;
  dummyStack.b.leaf.data = NULL;

#ifdef MSG_NON_REC
  msgnrstackinit(&msgStack);
#endif /* MSG_NON_REC */

  return TRUE;
}

/*
 * Utility functions
 */

static unsigned atomicInc(where)
    unsigned long *where;
{
  return (*where)++;
}

static unsigned atomicDec(where)
    unsigned long *where;
{
  xTrace1(msg, TR_FULL_TRACE, "atomicDec %lx", (u_long)where);
  xTrace1(msg, TR_FULL_TRACE, "atomicDec %lx", *where);
  return (*where)--;
}

static void incRef(this)
    MNode *this;
{
  xAssert(this != (MNode *)0 && this != (MNode *)-1);
  xAssert(this->type != t_MNodeJunk);
  xTrace2(msg, TR_MAJOR_EVENTS, "incRef %lx refCnt %d",
	  (u_long) this, this->refCnt);
 (void) atomicInc(&this->refCnt);
}

#define FREEONE(prefix, this) \
  xTrace1(msg, TR_MAJOR_EVENTS, "FREEONE dis= %lx", (u_long) this); \
  xFree((char *)this); \
  RECORD_RELEASE(prefix); \
  xTrace0(msg, TR_FULL_TRACE, "FREEONE done"); 


#ifndef MSG_NON_REC

/* 
 * decRef  - decrement reference count and free message if necessary
 * 
 * If the reference count reaches 0, the node is freed by the following 
 *  three cases:
 * 1. Recursively follow the b.pair structure and free the top node
 * 2. Free the page
 * 3. Use a node-specific function to free the b.buf.data structure and
 *    free the node structure
 */
static void decRef(this)
    MNode *this;
{
  xTrace1(msg, TR_FULL_TRACE, "decRef %lx", (u_long) this);
  xAssert(this != (MNode *)0 && this != (MNode *)-1 && this->refCnt >= 1);
  xTrace5(msg, TR_MAJOR_EVENTS,
	  "decRef %lx refCnt %ld addr %lx arg addr %lx type %d",
	  (u_long) this, this->refCnt, (u_long) decRef, (u_long) &this,
	  this->type);

  if (atomicDec(&this->refCnt) == 1) {
    switch(this->type) {
    case t_MNodePage:
      if ( this->b.page.size > MNodePageDataSize ) {
	  /* 
	   * Remove allocation for any extra pages attached to this node. 
	   */
	  msgSpace_used -= (this->b.page.size - MNodePageDataSize);
      }
      FREEONE(mpg, this);
      break;
    case t_MNodePair:
      decRef(this->b.pair.l.node);
      decRef(this->b.pair.r.node);
      FREEONE(mpr, this);
      break;
    case t_MNodeUsrPage:
      xTrace3(msg, TR_FULL_TRACE, "decRef t_MNodeUsrPage %lx %lx %ld",
	      (u_long) this->b.usrpage.bFree, (u_long) this->b.usrpage.data,
	      this->b.usrpage.size);
      this->b.usrpage.bFree(this->b.usrpage.data, this->b.usrpage.size);
      FREEONE(mup, this);
      break;
    case t_MNodeDummy:
      break;
    case t_MNodeBuf:
      xTrace2(msg, TR_FULL_TRACE, "decRef t_MNodeBuf %lx %lx",
	      (u_long) this->b.buf.bFree, (u_long) this->b.buf.data);
      this->b.buf.bFree(this->b.buf.data);
      FREEONE(mbf, this);
      break;
    default:
      printf("xkpanic: decRef wrong type %lx %d\n", (u_long) this, this->type);
      xAssert(0);
    }
  }
  xTrace1(msg, TR_FULL_TRACE, "decRef done %lx", (u_long) this);
}

#else

#if defined(__STDC__) || defined(__GNUC__)
static void msgnrpush_overflow(struct mstack *);
static void msgnrstackreinit(struct mstack *);
static bool msgnrNodeDestroy( struct mstack *, MNodePair *, void  *);
static void msgNodeForEach( MNode *, XNodeFun, XLeafFun, VOID *,  struct mstack *);
#else
static void msgnrpush_overflow();
static void msgnrstackreinit();
static bool msgnrNodeDestroy( );
static void msgNodeForEach( );
#endif

static bool decRefnr(this)
    MNode *this;
{
  xTrace1(msg, TR_FULL_TRACE, "decRefnr %x", this);
  xAssert(this != (MNode *)0 && this != (MNode *)-1 && this->refCnt >= 1);
  xTrace5(msg, TR_MAJOR_EVENTS, "decRefnr %x refCnt %d addr %x arg addr %x type %d", this, this->refCnt, decRefnr, &this, this->type);

  if (atomicDec(&this->refCnt) == 1) {
    switch(this->type) {
    case t_MNodePage:
      if ( this->b.page.size > MNodePageDataSize ) {
	  /* 
	   * Remove allocation for any extra pages attached to this node. 
	   */
	  msgSpace_used -= (this->b.page.size - MNodePageDataSize);
      }
      FREEONE(mpg, this);
      break;
    case t_MNodeUsrPage:
      xTrace3(msg, TR_FULL_TRACE, "decRefnr t_MNodeUsrPage %x %x %d", this->b.usrpage.bFree,
	      this->b.usrpage.data, this->b.usrpage.size);
      this->b.usrpage.bFree(this->b.usrpage.data, this->b.usrpage.size);
      FREEONE(mup, this);
      break;
    case t_MNodeDummy:
      break;
    case t_MNodePair:
      xTrace1(msg, TR_ERRORS, "decRefnr called with pair! %x", this);
      return FALSE;
    case t_MNodeBuf:
      xTrace2(msg, TR_FULL_TRACE, "decRefnr t_MNodeBuf %x %x", this->b.buf.bFree,
	      this->b.buf.data);
      this->b.buf.bFree(this->b.buf.data);
      FREEONE(mbf, this);
      break;
    default:
      printf("xkpanic: decRef wrong type %x %d\n", this, this->type);
      xAssert(0);
    }
  }
  xTrace1(msg, TR_FULL_TRACE, "decRefnonrec done %x", this);
  return TRUE;
}

static void decRef(this)
     MNode *this;
{
  xTrace1(msg, TR_FULL_TRACE, "msg: decRef (nonrec) %x", this);
  msgnrstackreinit(&msgStack);
  msgNodeForEach(this, msgnrNodeDestroy, decRefnr, 0, &msgStack);
}
#endif

#define PageMalloc(prefixsize, pages)  xMalloc(prefixsize * pages );

#define FINDAFREEONE(prefix, this, pages) { \
  this = (MNode *) PageMalloc(__CONCAT(prefix,size), pages); \
  if (this == (MNode *)0) { \
    Kabort("xkpanic: malloc failure in msgtool\n"); \
  } \
  this->refCnt = 1; \
  this->type = __CONCAT(prefix,type); \
  RECORD_USAGE(prefix);	\
  xTrace1(msg, TR_MAJOR_EVENTS, "FINDAFREEONE %lx", (u_long) this); \
}


/* creator/constructor */
static MNodePair *
newMNodePair(loff, llen, ln, roff, rlen, rn)
    long loff, llen, roff, rlen;
    MNode *ln, *rn;
{
  MNodePair *this;

  xTrace0(msg, TR_FULL_TRACE, "newMNodePair");
  FINDAFREEONE(mpr, this, 1);

  /* refCnt was set by the allocation above */
  this->b.pair.l.offset = loff;
  this->b.pair.l.length = llen;
  this->b.pair.l.node = ln;
  this->b.pair.r.offset = roff;
  this->b.pair.r.length = rlen;
  this->b.pair.r.node = rn;
  return this;
}



/* creator/constructor */
static MNodePage *
newMNodePage(pages)
    int	pages;
{
  MNodePage *this;

  xTrace0(msg, TR_FULL_TRACE, "newMNodePage");

#ifdef MSG_STATISTICS
  if ( pages > 1 ) {
      /* 
       * FINDAFREEONE takes care of statistics for single page
       * allocations, but we need to add the difference for larger
       * nodes. 
       */
      msgSpace_used += (pages - 1) * PageSize;
  }
#endif /* MSG_STATISTICS */

  FINDAFREEONE(mpg, this, pages);

  this->b.page.size = MNodePageDataSize + ((pages-1) * PageSize);
  this->b.page.data = this->b.page.buffer;

  return this;
}

/* creator/constructor */
static MNodePage *
newSmallMNodePage(bytes)
    int	bytes;
{
  MNodePage *this;

  xTrace1(msg, TR_FULL_TRACE, "newSmallMNodePage %d bytes", bytes);

#ifdef MSG_STATISTICS
  msgSpace_used += bytes;
#endif /* MSG_STATISTICS */

  this = (MNodePage *)xMalloc(bytes);
  this->refCnt = 1;
  this->type = mpgtype;
  RECORD_USAGE(mpg);
  xTrace1(msg, TR_MAJOR_EVENTS, "msgtool: small page %lx", (u_long) this); 

  this->b.page.size = bytes - MNodeLeafSize - KLUDGE;
  this->b.page.data = this->b.page.buffer;

  return this;
}

/* creator/constructor */
static MNodePage *
newMNodeUsrPage(freefunc, buffer, size)
     MsgCIPFreeFunc freefunc;
     char *buffer;
     long size;
{
  MNodePage *this;

  xTrace0(msg, TR_FULL_TRACE, "newMNodeUsrPage");
  FINDAFREEONE(mup, this, 1);

  this->b.usrpage.size = size;
  this->b.usrpage.data = buffer;
  this->b.usrpage.bFree = freefunc;
  return this;
}


/* construct an empty Msg						*/
void msgConstructEmpty(this)
    Msg *this;
{
  xTrace0(msg, TR_FULL_TRACE, "msgConstructEmpty");

  this->tree = this->stack = &dummyStack;
  incRef(this->tree);
  this->stackHeadPtr = this->stackTailPtr= this->headPtr = this->tailPtr = NULL;
  this->state.myStack = 0; /* so that push won't write */
  this->tailstate.myLastStack = 0;
  this->state.numNodes = 1;
  this->lastStack = 0;
  this->attr = 0;
  this->attrLen = 0;
}


/* construct a Msg with a copy of the content of the buffer buf with */
/* length len; this constructor causes data to be copied!! */
void msgConstructBuffer(this, buf, len)
    Msg *this;
    char *buf;
    long len;
{
  /* round up to nearest multiple of msg PageSize */
  int roundlen = ROUND4(len);

  xTrace0(msg, TR_FULL_TRACE, "msgConstructBuffer");

  xAssert(len > 0 ); /* use msgConstructEmpty instead */
  this->tree = this->stack = newMNodePage(PAGEROUND(roundlen));
  this->stackHeadPtr = this->stack->b.leaf.data + this->stack->b.leaf.size - roundlen;
  this->headPtr = this->stack->b.leaf.data;
  this->stackTailPtr = this->tailPtr = this->stackHeadPtr + len;
  this->state.myStack = 1;
  this->state.numNodes = 1;
  this->lastStack = this->stack;
  this->lastStackTailPtr = this->stackTailPtr;
  this->tailstate.myLastStack = 1;
  /* copy in data */
  bcopy(buf, this->stackHeadPtr, len);
  this->attr = 0;
  this->attrLen = 0;
}

/* construct a Msg that is a copy of another Msg			*/
void msgConstructCopy(this, another)
    Msg *this, *another;
{
  xTrace0(msg, TR_FULL_TRACE, "msgConstructCopy");
  incRef(another->tree);

  *this = *another;
  this->state.myStack = 0;
  this->tailstate.myLastStack = 0;
}

/* construct a Msg that refers to a contig				*/
void msgConstructContig(this, contig)
    Msg *this;
    MContig *contig;
{
  xTrace0(msg, TR_FULL_TRACE, "msgConstructContig");
  xAssert(0);
  /* This implementation assumed contig was an MNode, not an Mcontig.  menze */
#ifdef MCONTIG
  incRef(contig);
  this->offset = 0;
  this->length = contig->b.leaf.size;
  this->tree = contig;
  this->stack = contig;
  this->attr = 0;
  this->attrLen = 0;
#endif
}

/* constructor... to this Msg an uninitialized contiguous piece of memory
 * of length len and return a pointer to it.
 * Question:  Should this worry about making sure that the left edge of the
 * buffer is long word aligned?
 */
void msgConstructAllocate(this, len, buf)
    Msg *this;
    long len;
    char **buf;
{
  /* round up to nearest multiple of msg PageSize */
  int roundlen = ROUND4(len);

  xTrace0(msg, TR_FULL_TRACE, "msgConstructAllocate");
  xAssert(len > 0);
  if (roundlen+MNodeLeafSize+sizeof(int)+KLUDGE < PageSize/2) {
    int bytes = PageSize/2;
    while (bytes/2 > roundlen+MNodeLeafSize+sizeof(int)+KLUDGE) bytes /= 2;
    this->tree = this->stack = newSmallMNodePage(ROUND4(bytes-sizeof(int)));
    xTrace1(msg, TR_MAJOR_EVENTS, "msgConstructAllocate small page %d", roundlen);
  }
  else
    this->tree = this->stack = newMNodePage(PAGEROUND(roundlen));
  this->stackHeadPtr = this->stack->b.leaf.data + this->stack->b.leaf.size - roundlen;
  this->headPtr = this->stack->b.leaf.data;
  this->stackTailPtr = this->tailPtr = this->stackHeadPtr + len;
  this->state.myStack = 1;
  this->state.numNodes = 1;
  *buf = this->stackHeadPtr;
  this->lastStack = this->tree;
  this->lastStackTailPtr = this->stackTailPtr;
  this->tailstate.myLastStack = 1;
  this->attr = 0;
  this->attrLen = 0;
}


/*
 * msgConstructInplace()
 *
 *  make a data area into a message; the caller's routine is responsible
 *  for eventually freeing the data area, msgDestroy will only
 *  deallocate a small internal node buffer.  The caller must not
 *  directly deallocate the buffer; it must be a side-effect of
 *  msgDestroy.
 *
 */
void
msgConstructInplace(this, stack, length, freefunc)
     Msg  *this;
     char *stack;
     long  length;
     MsgCIPFreeFunc freefunc;
{
  xTrace0(msg, TR_FULL_TRACE, "msgConstructInplace");

  xAssert(length > 0 );
  this->tree = this->stack = newMNodeUsrPage(freefunc, stack, length);
  this->headPtr = this->stackHeadPtr = stack;
  this->stackTailPtr = this->tailPtr = this->stackHeadPtr + length;
  this->state.myStack = 1;
  this->state.numNodes = 1;
  this->attr = 0;
  this->attrLen = 0;
  this->lastStack = this->tree;
  this->lastStackTailPtr = this->stackTailPtr;
  this->tailstate.myLastStack = 1;
}

/*
 * msgConstructAppend()
 *
 *  Allocate a buffer and reserve it for appending.
 *  The stack head is at the low address instead of the high.
 *  This will cause an overflow on a msgPush operation, but not
 *  on a msgAppend operation.
 *  Return to the user a pointer to the beginning of the buffer.
 *
 */
void
msgConstructAppend(this, totalsize, bufferptr)
     Msg   *this;
     long   totalsize;
     char **bufferptr;
{
  int roundlen = ROUND4(totalsize);

  xTrace0(msg, TR_FULL_TRACE, "msgConstructAppend");

  xAssert(totalsize > 0 );
  this->tree = this->stack = newMNodePage(PAGEROUND(roundlen));
  this->stackHeadPtr = this->stack->b.leaf.data + this->stack->b.leaf.size - roundlen;
  this->headPtr = this->stack->b.leaf.data;
  this->stackTailPtr = this->tailPtr = this->stackHeadPtr;
  this->state.myStack = 1;
  this->state.numNodes = 1;
  this->attr = 0;
  this->attrLen = 0;
  this->lastStack = this->stack;
  this->lastStackTailPtr = this->stackTailPtr;
  this->tailstate.myLastStack = 1;
  *bufferptr = this->stackTailPtr;
}

/*
 *  msgAppend
 *
 *   copy data to the end of the last stack, if there is room;
 *     otherwise, allocate a new buffer and use msgJoin.
 *   NB the user supplies the size of the new message buffer
 *
 */
void
msgAppend(this, appendfunc, tail, tailLen, arg, newlength)
     Msg	*this;
     MStoreFun   appendfunc;
     VOID	*tail;
     VOID       *arg;
     long	 tailLen, newlength;
{
  char		*where;

  xTrace0(msg, TR_FULL_TRACE, "msgAppend");
  xAssert(tailLen >= 0 );

  if (this->tailstate.myLastStack &&
      this->lastStack &&
      ((this->lastStack == this->stack &&
	this->lastStackTailPtr >= this->stackTailPtr &&
	this->lastStackTailPtr + tailLen < this->lastStack->b.leaf.data + this->lastStack->b.leaf.size)
       ||
       (this->lastStackTailPtr + tailLen < this->lastStack->b.leaf.data + this->lastStack->b.leaf.size))) {
    /* cleanest case - there is room in the last stack */
      where = this->lastStackTailPtr;
      appendfunc(tail, where, tailLen, arg);    
      this->lastStackTailPtr += tailLen;
      this->stackTailPtr = this->lastStackTailPtr;
      this->tailPtr += tailLen;
    }
  else {
    /* this is more work; must get a new stack for the tail */
    Msg		newMsg;

    msgConstructAppend(&newMsg, newlength, &where);
    appendfunc(tail, where, tailLen, arg);    
    newMsg.stackTailPtr = (newMsg.tailPtr += tailLen);
    newMsg.lastStackTailPtr = newMsg.stackTailPtr;
    msgJoin(this, this, &newMsg);
    msgDestroy(&newMsg);
    /*
     * I lost last stack ownership after the join
     * I should regain it back because the newMsg is gone
     */
    this->tailstate.myLastStack = 1;
    return;
  }
}


/* assignment								*/
void msgAssign(this, another)
    Msg *this, *another;
{
  xTrace0(msg, TR_FULL_TRACE, "msgAssign");

  if (this != another) {
    incRef(another->tree);
    decRef(this->tree);
    *this = *another;
    this->state.myStack = 0;
    this->tailstate.myLastStack = 0;
  }
}


/* truncate this Msg to length newLength				*/
void msgTruncate(this, newLength)
     Msg *this;
     long newLength;
{
  long delta;

  xTrace0(msg, TR_FULL_TRACE, "msgTruncate");
  delta = this->tailPtr - this->stackHeadPtr - newLength;
  if (delta > 0)
    this->tailPtr -= delta; 
  if (this->lastStack == this->stack) {
    this->stackTailPtr = this->tailPtr;
    this->lastStackTailPtr = this->tailPtr;
  }
  else
    this->lastStack = 0;
}


/* remove a chunk of length len from the head of this Msg		*/
/* and assign it to head						*/
void msgChopOff(this, head, len)
     Msg *this, *head;
     long len;
{
  long mlen = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgChopOff");

  if (len < 0 || len > mlen) len = mlen;

  if (this != head) {
    incRef(this->tree);
    decRef(head->tree);
    head->stack = this->stack;
    head->tree = this->tree;

    head->headPtr = this->headPtr;
    head->stackHeadPtr = this->stackHeadPtr;
    head->stackTailPtr = this->stackTailPtr;
    head->state = this->state;
    this->state.myStack = 0;  /*  giving stack away */
    this->tailstate.myLastStack = 0;  /*  giving stack away */

    /* pop off the header */
    this->stackHeadPtr += len;
  }
  /* truncate the header */
  head->tailPtr = head->stackHeadPtr + len;
}

/* assign to this Msg the concatenation of Msg1 and Msg2		*/
void msgJoin(this, msg1, msg2)
  Msg *this, *msg1, *msg2;
{
  long msg1Length = msg1->tailPtr - msg1->stackHeadPtr;
  long msg2Length = msg2->tailPtr - msg2->stackHeadPtr;
  long msg2Offset;

  xTrace0(msg, TR_FULL_TRACE, "msgJoin");
  if (msg1Length == 0) {
    /* result is just msg2 */
    msgAssign(this, msg2);
 
  } else if (msg2Length == 0) {
    /* result is just msg1 */
    msgAssign(this, msg1);

  } else {
    /* create a new pair node */
    /* increment refCnts      */
    incRef(msg1->tree);
    incRef(msg2->tree);
    /* cleanup my old binding */
    decRef(this->tree);
    /* create a new pair node */
    msg2Offset = msg2->stackHeadPtr - msg2->headPtr;

    this->tree = newMNodePair(0L, msg1->tailPtr - msg1->headPtr, msg1->tree,
			      msg2Offset, msg2Length, msg2->tree);
    xAssert(this->tree != 0);
    this->headPtr = msg1->headPtr;
    this->tailPtr = msg1->tailPtr + msg2Length;
    this->stack = msg1->stack;
    this->stackHeadPtr = msg1->stackHeadPtr;
    this->stackTailPtr = msg1->stackTailPtr;
    this->state.myStack = (this == msg1) && this->state.myStack;
    this->state.numNodes = msg1->state.numNodes + msg2->state.numNodes + 1;
    this->tailstate.myLastStack = (this == msg2) && msg2->tailstate.myLastStack;
    this->lastStack = msg2->lastStack;
    this->lastStackTailPtr = msg2->lastStackTailPtr;
#ifdef CONTROL_RESOURCES
   if (this->state.numNodes > numNodesHardLimit) makeContiguous();
#endif /* CONTROL_RESOURCES */
  }
}

static char *msgPushOverflow(this, hdrLen)
    Msg *this;
    long hdrLen;
{
  long offset = this->stackHeadPtr - this->headPtr;
  long length = this->tailPtr - this->stackHeadPtr;
  long roundlen = ROUND4(hdrLen);

  xTrace0(msg, TR_FULL_TRACE, "msgPushOverflow");

  /* create a new root node and a new stack */
  this->stack = newMNodePage(PAGEROUND(roundlen));
  xAssert(this->stack != 0);
  this->tree =  newMNodePair(0L, this->stack->b.leaf.size, this->stack, offset, length, this->tree);
  xAssert(this->tree != 0);
  this->stackTailPtr = this->stack->b.leaf.data + this->stack->b.leaf.size;
  this->stackHeadPtr = this->stackTailPtr - hdrLen;
  this->headPtr = this->stack->b.leaf.data;
  this->tailPtr = this->stackTailPtr + length;
  this->state.myStack = 1;
  this->state.numNodes += 2;
  return this->stackHeadPtr;
}

/* push a header onto this Msg						*/
void msgPush(this, store, hdr, hdrLen, arg)
    Msg *this;
    MStoreFun store;
    VOID *hdr;
    long hdrLen;
    VOID *arg;
{
  char *where;

  xTrace0(msg, TR_FULL_TRACE, "msgPush");
  if ((this->state.myStack ||
       (this->tree == this->stack &&  /* see if we can cheaply own the stack */
        this->tree->refCnt == 1 &&
	(this->state.myStack = 1))) &&
      this->stackHeadPtr <= this->stackTailPtr &&
      this->stackHeadPtr - this->stack->b.leaf.data >= hdrLen) {
    this->stackHeadPtr -= hdrLen;
    where = this->stackHeadPtr;
  } else {
    /* need a new stack */
    where = msgPushOverflow(this, hdrLen);
  }
  store(hdr, where, hdrLen, arg);
}


/*
 * helper function called during foreach()
 * to copy the stack in topUnderflow()
 */
struct arg {
  char *buf;
  long size;
};
static bool
s_copy(buf, len, arg)
    char *buf;
    long len;
    VOID *arg;
{
  struct arg *a = (struct arg *)arg;
  long chunk_size;
  int chunki;

  xTrace0(msg, TR_FULL_TRACE, "s_copy");
  chunk_size = (len < a->size) ? len : a->size;
  /* bcopy takes an int, so be careful */
  chunki = chunk_size;
  xAssert( (long) chunki == chunk_size);
  bcopy(buf, a->buf, chunki);
  a->buf += chunk_size;
  a->size -= chunk_size;
  return (a->size != 0);
}


/* 
 * msgTopUnderflow - restructure a message with short initial node
 *
 * returns a pointer to the beginning of the data
 *
 * When the first node of a message has too little data to satisfy a
 * msgPop request, the message is restructured to make the requested
 * data reside in one stack.
 *
 * This is an inefficient operation that should be rarely called.
 */

static char *msgTopUnderflow(this, hdrLen)
    Msg *this;
    long hdrLen;
{
  struct arg c_args;
  MNodePage *newStack;
  int roundlen = ROUND4(hdrLen);

  xTrace0(msg, TR_FULL_TRACE, "msgTopUnderflow");
  xTrace0(msg, TR_SOFT_ERROR, "msgTopUnderflow");

  if (this->stackHeadPtr == this->stackTailPtr) {
    /* we are right at the boundary of the current stack */
    /* find the leftmost leaf node just to the right of the stack */
    MNodePair *newLeaf;
    MNodePair *newNode = (MNodePair*)(this->tree);
    long newOff = this->stackHeadPtr - this->headPtr;
    long newLen = 0;

    while (newNode->type == t_MNodePair) {
      if (newOff < newNode->b.pair.l.length) {
	/* explore the left subtree */
	newOff += newNode->b.pair.l.offset;
	newLen = newNode->b.pair.l.length;
	newNode = (MNodePair*)newNode->b.pair.l.node;
      }
      else {
	/* explore the right subtree */
	newOff = newOff - newNode->b.pair.l.length + newNode->b.pair.r.offset;
	newLen = newNode->b.pair.r.length;
	newNode = (MNodePair*)newNode->b.pair.r.node;
      } 
    }
    /* found a leaf node */
    newLeaf = (MNodeLeaf*)newNode;
    newLen -= newOff;

    /* printf("found a leaf node, newLen=%d\n", newLen); */
    /* check if the leaf node contains the entire requested head */
    if (newLen >= hdrLen) {
      long offset, length;

      /* just make the leaf node the new stack */
      this->stack = newLeaf;
      
      offset = this->stackHeadPtr - this->headPtr;
      length = this->tailPtr - this->stackHeadPtr;
      this->stackHeadPtr = this->stack->b.leaf.data + newOff;
      this->stackTailPtr = this->stackHeadPtr + newLen;
      this->headPtr = this->stackHeadPtr - offset;
      this->tailPtr = this->stackHeadPtr + length;
      this->state.myStack = 0; 
      return this->stackHeadPtr;
    }

    /* else we have to copy after all... */
  }
  xTrace0(msg, TR_SOFT_ERROR, "msgtopunderflow will copy");

  /* create a new root node and a new stack */
  newStack = newMNodePage(PAGEROUND(roundlen));
  xAssert(newStack != (MNodePage *)0);

  /* fill the new stack with the requested head */
  c_args.buf = newStack->b.leaf.data + newStack->b.leaf.size - hdrLen;
  c_args.size = hdrLen;
  /* overflow checks handled by msgForEach */
  msgForEach(this, s_copy, &c_args);

  /* assign new message */
  {
    long offset = this->stackHeadPtr - this->headPtr;
    long length = this->tailPtr - this->stackHeadPtr;

    this->tree = newMNodePair(0L, newStack->b.leaf.size, newStack,
			      offset + hdrLen, length - hdrLen, this->tree);
    xAssert(this->tree != 0);
    this->stack = newStack;
    this->stackTailPtr = this->stack->b.leaf.data + this->stack->b.leaf.size;
    this->stackHeadPtr = this->stackTailPtr - hdrLen;
    this->headPtr = this->stack->b.leaf.data;
    this->tailPtr = this->stackHeadPtr + length;
    this->state.myStack = 1;
    this->state.numNodes += 2;
    return this->stackHeadPtr;
  }
}


/* msgPop
 *
 * pop a header from this Msg
 * returns 1 after successful pop
 *   
 * calls msgTopUnderflow if necessary
 */
bool msgPop(this, load, hdr, hdrLen, arg)
    Msg *this;
    MLoadFun load;
    VOID *hdr;
    long hdrLen;
    VOID *arg;
{
  char *where;
  long actualLen;
  long length = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgPop");
  xAssert(hdrLen>=0);
  if (!hdrLen) xTrace0(msg, TR_SOFT_ERROR, "msgPop zero length - useless");
  if (hdrLen <= length) {
    /* Msg is long enough */
    if (this->stackHeadPtr + hdrLen <= this->stackTailPtr) {
      where = this->stackHeadPtr;
    }
    else {
      where = msgTopUnderflow(this, hdrLen);
    }
    actualLen = load(hdr, where, hdrLen, arg); 
    xAssert(actualLen >= 0 && actualLen <= hdrLen);
    if (actualLen > 0) {  /* if a "peek", preserve ownerships */
      this->stackHeadPtr += actualLen;
      this->state.myStack = 0;

      /* quick check to see if we can still keep last stack ownerships */
      if (this->lastStack == this->stack)
	this->tailstate.myLastStack = 0;
    }
    return TRUE;
  }
  return FALSE;
}


/* msgPopDiscard
 *
 * pop and discard an object of length len
 * returns TRUE after successful pop
 */
bool msgPopDiscard(this, len)
    Msg *this;
    long len;
{
  long length;

  xTrace0(msg, TR_FULL_TRACE, "msgPopDiscard");
  xAssert(len>=0);
  xIfTrace(msg, TR_SOFT_ERROR) {
    if (len==0)
      printf("msgPopDiscard of length 0; useless call");
  }
  length = this->tailPtr - this->stackHeadPtr;
  if (len > length) len = length;
  this->stackHeadPtr += len;
  this->state.myStack = 0;
  this->tailstate.myLastStack = 0;
  return TRUE;
}



/* import a Msg from a different address space				*/
/* makes the content of the Msg foreign (contigs which may reside in	*/
/* a different address space) accessible in the current address space*/
/*ARGSUSED*/
void msgImport(this, foreign)
    Msg *this, *foreign;
{
  xTrace0(msg, TR_FULL_TRACE, "msgImport");
}


/* lock the memory associated with the Msg in physical memory		*/
/* returns 0, if not all contigs of the Msg can be locked		*/
/*ARGSUSED*/
bool msgLock(this)
    Msg *this;
{
  xTrace0(msg, TR_FULL_TRACE, "msgLock");
  return 1;
}


/* unlock the memory associated with the Msg				*/
/* returns 0, if not all contigs of the Msg can be unlocked		*/
/*ARGSUSED*/
bool msgUnlock(this)
    Msg *this;
{
  xTrace0(msg, TR_FULL_TRACE, "msgUnlock");
  return 1;
}


xkern_return_t
msgSetAttr( this, name, attr, len )
    Msg		*this;
    int		name;
    VOID	*attr;
    int		len;
{
    /* 
     * Only the default attribute (name == 0) is supported at this time. 
     */
    if ( name != 0 ) {
	xTrace1(msg, TR_SOFT_ERRORS,
		"msgSetAttr called with unsupported name %d", name);
	return XK_FAILURE;
    }
#ifdef XK_DEBUG
    if ( this->attr ) {
	xTrace0(msg, TR_SOFT_ERRORS,
		"msgSetAttr overriding previous non-null attribute");
    }
#endif
    this->attr = attr;
    this->attrLen = len;
    return XK_SUCCESS;
}


VOID *
msgGetAttr( this, name )
    Msg		*this;
    int		name;
{
    /* 
     * Only the default attribute (name == 0) is supported at this time. 
     */
    if ( name != 0 ) {
	xTrace1(msg, TR_SOFT_ERRORS,
		"msgGetAttr called with unsupported name %d", name);
	return 0;
    }
    return this->attr;
}

/*
 * Forward:
 */
static void	msgPairShow( MNode *, long, long, long );
static void	msgLeafShow( MNode *, long, long, long );

static char blanks[] = "                                ";

#ifdef MSG_NON_REC
/*
 *  Message Iteration Functions
 *
 *  Routines for processing the message data; hides the message tree
 *   structure and lengths of node buffers
 */

/*
 *  We use a single stack for processing message nodes: msgStack
 *
 *  MP-safe versions will have to put a lock on getting the
 *  msgStack.  Probably will have a pool of stacks.
 *
 */
static void msgnrstackinit(stack)
     struct mstack *stack;
{
  xTrace0(msg, TR_FULL_TRACE, "msgnrstackinit");

  stack->size = MSG_NR_STK_SZ;
  stack->top = stack->bottom = (MNodePage **)xMalloc(MSG_NR_STK_SZ);
  stack->limit = (MNodePage **)(((char *)(stack->bottom)) + MSG_NR_STK_SZ);
}

static void msgnrpush_overflow(stack)
     struct mstack *stack;
{
  char *newstack;
  int newsize = 2 * (stack->size +4) -4;

  xTrace0(msg, TR_FULL_TRACE, "msgnrpush_overflow");

  xAssert (stack->top < stack->limit);

  newstack = realloc(stack->bottom, newsize);
  if (newstack != (char *)(stack->bottom)) {
    bcopy((char *)(stack->bottom), newstack, stack->size);
    xFree((char *)stack->bottom);
    stack->top = (MNodePage **)(newstack + ((char *)stack->top - (char *)stack->bottom));
    stack->bottom = (MNodePage **)newstack;
  }
  stack->size = newsize;
  stack->limit = (MNodePage **)(((char *)(stack->bottom)) + stack->size);
  xTrace1(msg, TR_SOFT_ERROR, "msgnrpush overflow to %d", stack->size);
}

static void msgNodeForEach(this, fnode, fleaf, arg, stack)
    MNode *this;
    XNodeFun fnode;
    XLeafFun fleaf;
    VOID *arg;
    struct mstack *stack;
{
  bool stackcontinue = TRUE;
  bool nodecontinue = TRUE;

  xTrace0(msg, TR_FULL_TRACE, "msgNodeForEach");

  while (stackcontinue==TRUE && nodecontinue==TRUE) {
    switch(this->type) {
    case t_MNodePage:
      (void)( (fleaf)(this));
      break;
    case t_MNodePair:
      nodecontinue = (fnode)(stack, this, arg);
      break;
    case t_MNodeUsrPage:
      (void)( (fleaf)(this));
      break;
    case t_MNodeDummy:
      break;
    case t_MNodeBuf:
      (void)( (fleaf)(this));
      break;
    default:
      xAssert(0);
    }
    stackcontinue = msgnrpop(stack, &this);
    xAssert(this);
  }
}

/* 
 * msgnrNodeDestroy  - decrement reference count and free message if necessary
 * 
 * If the reference count reaches 0, the node is freed
 */
static bool msgnrNodeDestroy(stack, this, arg)
    struct mstack *stack;
    MNodePair *this;
    void  *arg;
{
  xTrace1(msg, TR_FULL_TRACE, "msgnrNodeDestroy %x", this);
  xAssert(this != (MNode *)0 && this != (MNode *)-1 && this->refCnt >= 1);
  xTrace5(msg, TR_MAJOR_EVENTS, "msgnrNodeDestroy %x refCnt %d addr %x arg addr %x type %d", this, this->refCnt, decRef, &this, this->type);

  if (atomicDec(&this->refCnt) == 1)
    if (this->type == t_MNodePair) {
      xAssert(this->b.pair.l.node);
      msgnrpush(stack, this->b.pair.l.node);
      xAssert(this->b.pair.r.node);
      msgnrpush(stack, this->b.pair.r.node);
      FREEONE(mpr, this);
    }
  xTrace1(msg, TR_EVENTS, "msgnrNodeDestroy done %x", this);
  return TRUE;
}

void msgDestroy(this)
     Msg *this;
{
  xTrace0(msg, TR_FULL_TRACE, "msgDestroy non rec");
  msgnrstackreinit(&msgStack);
  msgNodeForEach(this->tree, msgnrNodeDestroy, decRefnr, 0, &msgStack);
}

#ifdef MSG_NR_TEST

bool msgnrUserFun(this, off, len, arg)
     MNode *this;
     long off, len;
     VOID *arg;
{
  printf("Msg User func: %x %x %ld %ld\n", this, arg, off, len);
  return TRUE;
}

struct Uargs {
  long a, b; } UserArgs;

void msgnrtest()
{
  Msg m1, m2;
  char *buf;
  int i=0;

  msgConstructAllocate(&m1, 1024, &buf);
  while (i++ < 1000) {
    msgConstructAllocate(&m2, 1024, &buf);
    if (i < 10)  msgShow(&m1);
    /* if ( !(i % 100) ) msgShow(&m1); */
    msgJoin(&m1, &m2, &m1);
  }
  msgShow(&m1);
  msgForEach(&m1, msgnrUserFun, &UserArgs);
  msgDestroy(&m1);
}

#endif /* MSG_NR_TEST */

/*
 * Forward:
 */
/* 
 * types of user functions called in ForEach()
 */
#if defined(__STDC__) || defined(__GNUC__)
typedef bool (*XNodeFunArgs)( struct mstack *, MNode *, long, long );
typedef bool (*XLeafFunArgs)( struct mstack *, MNode *, long, long);
static bool	msgnrPairArgs( struct mstack *, MNodePair *, long, long );
static bool	msgnrLeafForEach( struct mstack *, MNodePair *, long, long );
static bool	msgnrLeafShow( struct mstack *, MNode *, long, long );
static void	msgNonRecForEach(struct mstack *, MNode *, XNodeFunArgs, XLeafFunArgs, long, long );

#else

typedef bool (*XNodeFunArgs)();
typedef bool (*XLeafFunArgs)();
static bool	msgnrPairArgs( );
static bool	msgnrLeafForEach( );
static bool	msgnrLeafShow( );
static void	msgNonRecForEach( );
#endif


static void msgnrstackreinit(stack)
     struct mstack *stack;
{
  xTrace0(msg, TR_FULL_TRACE, "msgnrstackreinit");

  stack->top = stack->bottom;
  stack->arg = 0;
  stack->indent = 0;
  stack->func = 0;
}

static bool msgnrPairArgs(stack, this, off, len)
    MNodePair *this;
    struct mstack *stack;
    long off, len;
{
  long chunk_size, lchunk=0, loff=0;
  struct nlink *k, *l;
  int indent = stack->indent;

  xTrace0(msg, TR_FULL_TRACE, "msgnrPairArgs");
  if (indent > sizeof(blanks)) indent = sizeof(blanks);
  xTrace5(msg, TR_FULL_TRACE, "%.*sPair: refCnt = %d off = %d len = %d\n",
	  indent, blanks, this->refCnt, off, len);
  /* left child */
  l = &this->b.pair.l;
  chunk_size = (off < l->length) ? l->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
    loff = l->offset + off;
    lchunk = chunk_size;
    len -= chunk_size;
  }
  
  off = off - l->length + chunk_size;

  /* right child */
  k = &this->b.pair.r;
  chunk_size = (off < k->length) ? k->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
      msgnrpush(stack, k->node);
      msgnrpush(stack, (MNode *)(k->offset+off));
      msgnrpush(stack, (MNode *)chunk_size);
    }
  if (lchunk) {
    msgnrpush(stack, l->node);
    msgnrpush(stack, (MNode *)loff);
    msgnrpush(stack, (MNode *)lchunk);
  }
  stack->indent += 2;
  return TRUE;
}

static bool msgnrLeafShow(stack, this, off, len)
    struct mstack *stack;
    MNode *this;
    long off, len;
{
  int lim, i;
  char *c;
  int indent = stack->indent;
  
  xTrace0(msg, TR_FULL_TRACE, "msgnrLeafShow");
  xAssert(this->b.leaf.size - off >= len);
  if (indent > sizeof(blanks)) indent = sizeof(blanks);
  printf("%.*sLeaf: %#x refCnt %d off %d len %d", indent, blanks, 
	 this->b.leaf.data, this->refCnt, off, len);
  /*
   * Display first 8 characters of data
   */
  printf("\tdata: ");
  lim = len < 8 ? len : 8;
  for (c=this->b.leaf.data+off, i=0; i < lim; c++, i++) {
    printf("%.2x ", *c & 0xff);
  }
  putchar('\n');
  stack->indent -= 2;
  return TRUE;
}

static void msgNonRecForEach(stack, this, fnode, fleaf, arg1, arg2)
    struct mstack *stack;
    MNode *this;
    XNodeFunArgs fnode;
    XLeafFunArgs fleaf;
    long arg1, arg2;
{
  bool stackcontinue = TRUE;
  bool nodecontinue = TRUE;

  xTrace0(msg, TR_FULL_TRACE, "msgNonRecForEach");

  while (stackcontinue==TRUE && nodecontinue==TRUE) {
    switch(this->type) {
    case t_MNodePage:
      (void)( (fleaf)(stack, this, arg1, arg2));
      break;
    case t_MNodePair:
      nodecontinue = (fnode)(stack, this, arg1, arg2);
      break;
    case t_MNodeUsrPage:
      (void)( (fleaf)(stack, this, arg1, arg2));
      break;
    case t_MNodeDummy:
      break;
    case t_MNodeBuf:
      (void)( (fleaf)(stack, this, arg1, arg2));
      break;
    default:
      xAssert(0);
    }
    stackcontinue = msgnrpop(stack, &arg2);
    stackcontinue = msgnrpop(stack, &arg1);
    stackcontinue = msgnrpop(stack, &this);
    xAssert(this);
  }
}

void msgShow(this)
    Msg *this;
{
  long offset = this->stackHeadPtr - this->headPtr;
  long length = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgShow nonrec");
  printf("Msg: Stack=%#x, stackHeadPtr=%#x, stackTailPtr=%#x, myStack=%d, myLastStack %d, numNodes=%d offset %ld length %ld\n",
	 this->stack, this->stackHeadPtr, this->stackTailPtr,
	 this->state.myStack, this->tailstate.myLastStack,
	 this->state.numNodes,
	 offset, length);
  msgnrstackreinit(&msgStack);
  if (length > 0)
    (void) msgNonRecForEach(&msgStack, this->tree, msgnrPairArgs, msgnrLeafShow, offset, length);
}

static bool msgnrLeafForEach(stack, this, off, len)
     struct mstack *stack;
     MNode *this;
     long off, len;
{
  xTrace0(msg, TR_FULL_TRACE, "msgnrLeafForEach");
  xAssert(this->b.leaf.size - off >= len);
  return (stack->func)(this->b.leaf.data + off, len, stack->arg);
}

/* for every contig in this Msg, invoke the function f with 		*/
/* arguments const char *buf (=address of contig), long len 		*/
/* (=length of contig), and VOID * arg (=user-supplied argument),  	*/
/* while f returns TRUE							*/
void msgForEach(this, f, arg)
    Msg *this;
    XCharFun f;
    VOID *arg;
{
  long offset = this->stackHeadPtr - this->headPtr;
  long length = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgForEach nonrec");

  /* try the common case of one leaf */
  if (this->tree->type == t_MNodePage || this->tree->type == t_MNodeUsrPage) {
    (f)(this->tree->b.leaf.data + offset, length, arg);
    return;
  }

  msgnrstackreinit(&msgStack);
  msgStack.func = f;
  msgStack.arg = arg;
  (void) msgNonRecForEach(&msgStack, this->tree, msgnrPairArgs, msgnrLeafForEach, offset, length);
}


#else

/* destructor								*/
void msgDestroy(this)
    Msg *this;
{
  xTrace1(msg, TR_FULL_TRACE, "msgDestroy arg this: %lx", (u_long) this);
  decRef(this->tree);
  xTrace0(msg, TR_FULL_TRACE, "msgDestroy done");
}

void msgShow(this)
    Msg *this;
{
  long offset = this->stackHeadPtr - this->headPtr;
  long length = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgShow");
  printf("Msg: Stack=%#lx, stackHeadPtr=%#lx, stackTailPtr=%#lx, myStack=%d, myLastStack %d, numNodes=%d offset %ld length %ld\n",
	 (u_long) this->stack, (u_long) this->stackHeadPtr,
	 (u_long) this->stackTailPtr,
	 this->state.myStack, this->tailstate.myLastStack,
	 this->state.numNodes,
	 offset, length);
  if (length > 0)
    (void) msgInternalShow(this->tree, offset, length, 0L);
}

/* for every contig in this Msg, invoke the function f with 		*/
/* arguments const char *buf (=address of contig), long len 		*/
/* (=length of contig), and VOID * arg (=user-supplied argument),  	*/
/* while f returns TRUE							*/

void msgForEach(this, f, arg)
    Msg *this;
    XCharFun f;
    VOID *arg;
{
  long offset = this->stackHeadPtr - this->headPtr;
  long length = this->tailPtr - this->stackHeadPtr;

  xTrace0(msg, TR_FULL_TRACE, "msgForEach");
  (void) msgInternalForEach(this->tree, offset, length, f, arg);
}


#endif /* MSG_NON_REC */

static bool msgPairForEach( MNodePair *, long, long, XCharFun, VOID * );
static bool msgLeafForEach( MNodePair *, long, long, XCharFun, VOID * );

static bool msgInternalForEach(this, off, len, f, arg)
    MNode *this;
    long off, len;
    XCharFun f;
    VOID *arg;
{
  xTrace0(msg, TR_FULL_TRACE, "msgInternalForEach");
  switch(this->type) {
    case t_MNodePage:
      return msgLeafForEach(this, off, len, f, arg);
    case t_MNodePair:
      return msgPairForEach(this, off, len, f, arg);
    case t_MNodeUsrPage:
      return msgLeafForEach(this, off, len, f, arg);
    case t_MNodeDummy: return TRUE;
    case t_MNodeBuf:
      return msgLeafForEach(this, off, len, f, arg);
    default:
      xAssert(0);
  }
  /*
   * To make the compiler happy
   */
  return FALSE;
}

static bool msgPairForEach(this, off, len, f, arg)
    MNodePair *this;
    long off, len;
    XCharFun f;
    VOID *arg;
{
  long chunk_size;
  struct nlink *k;

  xTrace0(msg, TR_FULL_TRACE, "msgPairForEach");
  /* left child */
  k = &this->b.pair.l;
  chunk_size = (off < k->length) ? k->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
    if (!msgInternalForEach(k->node, k->offset + off, chunk_size, f, arg))
      return 0;
    len -= chunk_size;
  }
  off = off - k->length + chunk_size;

  /* right child */
  k = &this->b.pair.r;
  chunk_size = (off < k->length) ? k->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
    if (!msgInternalForEach(k->node, k->offset + off, chunk_size, f, arg))
      return 0;
  }
  return 1;
}

static bool msgLeafForEach(this, off, len, f, arg)
    MNode *this;
    long off, len;
    XCharFun f;
    VOID *arg;
{
  xTrace0(msg, TR_FULL_TRACE, "msgLeafForEach");
  xAssert(this->b.leaf.size - off >= len);
  return f(this->b.leaf.data + off, len, arg);
}

static void msgInternalShow(this, off, len, indent)
    MNode *this;
    long off, len, indent;
{
    xTrace0(msg, TR_FULL_TRACE, "msgInternalShow");
    switch(this->type) {
      case t_MNodePair:
	msgPairShow(this, off, len, indent);
	break;
      case t_MNodePage:
      case t_MNodeUsrPage:
      case t_MNodeBuf:
	msgLeafShow(this, off, len, indent);
	break;
      default:
	break;
    }
}

static void msgPairShow(this, off, len, indent)
    MNodePair *this;
    long off, len, indent;
{
  long chunk_size;
  struct nlink *k;

  xTrace0(msg, TR_FULL_TRACE, "msgPairShow");
  printf("%.*sPair: refCnt = %ld off = %ld len = %ld\n", (int) indent, blanks, 
	 this->refCnt, off, len);
  /* left child */
  k = &this->b.pair.l;
  chunk_size = (off < k->length) ? k->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
    msgInternalShow(k->node, k->offset + off, chunk_size, indent+2);
    len -= chunk_size;
  }
  off = off - k->length + chunk_size;

  /* right child */
  k = &this->b.pair.r;
  chunk_size = (off < k->length) ? k->length - off : 0;
  if (chunk_size > len) chunk_size  = len;
  if (chunk_size) {
    msgInternalShow(k->node, k->offset + off, chunk_size, indent+2);
  }
}

static void msgLeafShow(this, off, len, indent)
    MNode *this;
    long off, len, indent;
{
  int lim, i;
  char *c;
  
  xTrace0(msg, TR_FULL_TRACE, "msgLeafShow");
  xAssert(this->b.leaf.size - off >= len);
  printf("%.*sLeaf: %#lx refCnt %ld off %ld len %ld", (int) indent, blanks, 
	 (u_long) this->b.leaf.data, this->refCnt, off, len);
  /*
   * Display first 8 characters of data
   */
  printf("\tdata: ");
  lim = len < 8 ? len : 8;
  for (c=this->b.leaf.data+off, i=0; i < lim; c++, i++) {
    printf("%.2x ", *c & 0xff);
  }
  putchar('\n');
}

/* copy this entire Msg into contiguous storage */
static void
makeContiguous(this)
     Msg *this;
{
  long length;
  struct arg c_args;
  Msg empty, tmp;

  xTrace0(msg, TR_FULL_TRACE, "makeContiguous");

  length = this->tailPtr - this->stackHeadPtr;

  if (length == 0) {
    /* free the Msg's resources */
    msgConstructEmpty(&empty);
    msgAssign(this, &empty);
    return;
  }

  if (this->stack == this->tree) 
    /* this Msg is contiguous */
    return;

  /* create a new Msg with a contiguous buffer */
  msgConstructAllocate(&tmp, length, &c_args.buf);

  /* copy into the new Msg */
  c_args.size = length;
  msgForEach(this, s_copy, &c_args);

  /* assign the new Msg */
  msgAssign(this, &tmp);
}

/* perform housecleaning to free unnecessary resources allocated to this msg */
void msgCleanUp(this)
     Msg *this;
{
#ifdef undef
  long offset = this->stackHeadPtr - this->headPtr;
#endif
  long length = this->tailPtr - this->stackHeadPtr;

  if (length > 0) {
    if (this->state.numNodes > numNodesSoftLimit) makeContiguous(this);

#ifdef undef
    /* prune the tree */
    MNode *oldTree = tree;

    treePrune(offset, length, &offset, tree);
    
    if (tree != oldTree) {
      /* the root node was pruned */
      /* must change headPtr to reflect the new offset */
      this->headPtr = this->stackHeadPtr - offset;
    }

    /* if the stack was empty, it may have been deallocated */
    /* set it not-owned, so push won't try to write into it */
    this->state.myStack = 0;
    this->tailstate.myLastStack = 0;
#endif /* undef */

  } else {
    /* this Msg is empty */
    /* free its resources */
    Msg empty;
    msgConstructEmpty(&empty);
    msgAssign(this, &empty);
  }
}


static bool
frag2Buf( frag, len, bufPtr )
    char	*frag;
    long	len;
    VOID 	*bufPtr;
{
    xTrace3(msg, TR_FUNCTIONAL_TRACE,
	    "frag2Buf copying %d bytes from %lx to %lx",
	    len, (u_long)frag, (u_long)(*(char **)bufPtr));
    bcopy(frag, *(char **)bufPtr, len);
    *(char **)bufPtr += len;
    return TRUE;
}


void
msg2buf( msg, buf )
    Msg		*msg;
    char	*buf;
{
    char	*ptr = buf;

    xTrace2(msg, TR_FUNCTIONAL_TRACE, "msg2Buf (%lx, %lx)",
	    (u_long) msg, (u_long) buf);
    msgForEach(msg, frag2Buf, &ptr);
    xTrace0(msg, TR_FULL_TRACE, "msg2Buf returns");
}


void
msgStats()
{
    xTrace0(msg, TR_ALWAYS,
	    "x-kernel message statistics (current, high water):");
    xTrace2(msg, TR_ALWAYS, "    Pair nodes:      %6d, %6d",
	    mpr_used, mpr_hiwat);
    xTrace2(msg, TR_ALWAYS, "    Page nodes:      %6d, %6d",
	    mpg_used, mpg_hiwat);
    xTrace2(msg, TR_ALWAYS, "    User Page nodes: %6d, %6d",
	    mup_used, mup_hiwat);
    xTrace2(msg, TR_ALWAYS, "    Buffer nodes:    %6d, %6d",
	    mbf_used, mbf_hiwat);
    xTrace2(msg, TR_ALWAYS, "Total space used: %9d, %9d",
	    msgSpace_used, msgSpace_hiwat);
}
