/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_thread.c,v $
 *
 * HISTORY
 * $Log: scout_thread.c,v $
 * Revision 1.3  1994/06/11  00:02:32  menze
 * Added copyright notice
 *
 * Revision 1.2  1994/05/14  23:56:26  davidm
 * (show_blocked): added.  Helps to keep track of where your threads go.
 *
 * Revision 1.1  1994/04/15  17:47:46  davidm
 * Initial revision
 */
/*
 * Threads package.
 */
#include "scout_assert.h"
#include "scout_fifo.h"
#include "scout_malloc.h"
#include "scout_trace.h"
#include "scout_thread.h"
#include "scout_thread_arch.h"
#include "scout_util.h"

#define NPRIORITIES	(SCOUT_THREAD_PRIORITY_MIN \
			 - SCOUT_THREAD_PRIORITY_MAX + 1)

int trace_scout_thread = 0;

static void start (void)  NORETURN;	/* forward declaration */
static void retire (void) NORETURN;	/* never returns */

static Scout_Thread_Options def_opts =
{
    "noname",
    0,
    DEFAULT_STACK_SIZE,
    SCOUT_THREAD_PRIORITY_STD
};
const Scout_Thread_Options	*scout_thread_default_options = &def_opts;

/*
 * Invariant: the currently running thread is "scout_thread_self" (it is
 * *not* in any ready_q!).
 */
Scout_Thread			*scout_thread_self = 0;

static Scout_Thread	*dead_man = 0;
static u_int		high_prio = SCOUT_THREAD_PRIORITY_MAX;
static Scout_Queue	ready_q[NPRIORITIES];
static Scout_FIFO	async_q;

#ifdef SCOUT_DEBUG
# define STACK_CHECK() \
{ \
    if (scout_thread_free_stack() <= 0 || \
	scout_thread_self->magic != MAGIC_COOKIE) \
    { \
	error("Stack Overflow!"); \
	retire(); \
    } /* if */ \
}
#else
# define STACK_CHECK()
#endif /* SCOUT_DEBUG */


#ifdef SCOUT_DEBUG

#define MAX_THREADS	256

static Scout_Thread	*blocked[MAX_THREADS];

#define HASH(a)	(((u_long)(a) >> 3) % MAX_THREADS)


static void
mark_blocked(Scout_Thread *t)
{
    int i, i0;

    i = i0 = HASH(t);
    while (blocked[i]) {
	i = (i + 1) % MAX_THREADS;
	if (i == i0) {
	    scout_panic("scout_thread: mark_blocked: too many threads!");
	} /* if */
    } /* while */
    blocked[i] = t;
} /* mark_blocked */


static void
mark_unblocked(Scout_Thread *t)
{
    int i, i0;

    i = i0 = HASH(t);
    while (blocked[i] != t) {
	i = (i + 1) % MAX_THREADS;
	if (i == i0) {
	    scout_panic("scout_thread: mark_unblocked: couldn't find thread!");
	} /* if */
    } /* while */
    blocked[i] = 0;
} /* mark_unblocked */


void
show_blocked(void)
{
    Scout_Thread *t;
    int i;

    printf("%-16s %-16s %-16s name:\n",
	   "tcb address:", "sp:", "pc:");
    for (i = 0; i < MAX_THREADS; ++i) {
	t = blocked[i];
	if (t) {
	    printf("%16p %16p %16p %s\n",
		   t, t->stack_pointer,
		   t->running ? GET_SAVED_PC(t->stack_pointer) : t->fun,
		   scout_thread_get_name(t));
	} /* if */
    } /* for */
} /* show_blocked */

#endif /* SCOUT_DEBUG */


/*
 * Print string using minimal amount of stack space:
 */
static void
error(const char *msg)
{
    int ch;

    while ((ch = *msg++) != '\0') {
	putchar(ch);
    } /* while */
    putchar('\n');
} /* error */


static void
idler(void *arg)
{
    while (TRUE) {
	scout_thread_yield();
    } /* while */
} /* idler */


/*
 * Switch context from scout_thread_self to t.  This is tricky code.  You
 * should verify the generated assembly code.
 */
static void
cswtch(Scout_Thread *t)
{
    static Scout_Thread *old;

    old = scout_thread_self;
    scout_thread_self = t;
    if (old) {
	if (old->fpu_dirty) {
	    old->fpu_dirty = FALSE;
	    SAVE_FPU_STATE(old->fpu_state);
	} /* if */
	SAVE_STATE(old->stack_pointer);
    } /* if */
    if (scout_thread_self->running) {
	LOAD_STATE(scout_thread_self->stack_pointer);
    } else {
	scout_thread_self->running = TRUE;
	/*
	 * Switch to new stack and invoke start(); the second argument
	 * is just there to tell GCC that start() is really used.
	 */
	SWITCH_STACK_AND_START(scout_thread_self->stack_pointer, start);
    } /* if */
} /* cswtch */


static void
sched(void)
{
    Scout_Thread *next;

    /* now is the time move asynchronously awakened threads: */
    while ((next = scout_fifo_remove(&async_q)) != 0) {
#ifdef SCOUT_DEBUG
	mark_unblocked(next);
#endif
	scout_assert(next->option.priority <= SCOUT_THREAD_PRIORITY_MIN);
	scout_queue_append(&ready_q[next->option.priority], next);
	/* maintain high_prio hint: */
	if (next->option.priority < high_prio) {
	    high_prio = next->option.priority;
	} /* if */
    } /* while */

    /* dispatch to highest priority thread: */
    while (1) {
	scout_assert(high_prio <= SCOUT_THREAD_PRIORITY_MIN);
	next = (Scout_Thread*) scout_queue_remove(&ready_q[high_prio]);
	if (next) {
	    cswtch(next);
	    if (dead_man) {
		ScoutTrace(scout_thread, TR_EVENTS,
			   (__FILE__": "
			    "freeing space of dead thread `%s' at %p",
			    scout_thread_get_name(dead_man), dead_man));
		scout_free(dead_man);
		dead_man = 0;
	    } /* if */
	    return;
	} /* if */
	++high_prio;
    } /* while */
} /* sched */


static Scout_Thread*
create(Scout_Thread_Func fun, const Scout_Thread_Options *options)
{
    Scout_Thread *t;
    char *mem;
    int name_len;
    u_int stack_size;

    /* compute size of name and round up to multiple of sizeof(void*): */
    name_len = strlen(options->name) + 1 + sizeof(void*) - 1;
    name_len &= ~(sizeof(void*) - 1);

    /*
     * Make sure stack is big enough to hold state plus initial
     * frame:
     */
    stack_size = options->stack_size & ~(sizeof(void*) - 1);
    if (stack_size < MIN_USEFUL_STACK_SIZE) {
	ScoutTrace(scout_thread, TR_ERRORS,
		   (__FILE__": "
		    "stack too small (%d bytes, should be >%d bytes)",
		    stack_size, MIN_USEFUL_STACK_SIZE));
	return 0;
    } /* if */
    if (options->priority > SCOUT_THREAD_PRIORITY_MIN) {
	ScoutTrace(scout_thread, TR_ERRORS,
		   (__FILE__": illegal thread priority %d specified",
		    options->priority));
	return 0;
    } /* if */

    /*
     * Allocate thread descriptor, name string, FPU state, and stack
     * in one chunk:
     */
    mem = scout_malloc(stack_size);
    if (!mem) {
	ScoutTrace(scout_thread, TR_EVENTS,
		   (__FILE__": not enough memory for new thread"));
	return 0;
    } /* if */

    /* initialize memory: */
    t = (Scout_Thread*) mem;
    t->option		= *options;
    t->stack_high	= mem + stack_size - sizeof(void*);
    t->fpu_state	= mem + sizeof(Scout_Thread);
    t->option.name	= (char*) t->fpu_state + FPU_STATE_SIZE;
    t->running		= FALSE;
    t->fpu_dirty	= FALSE;
    t->fun		= fun;
    t->magic		= MAGIC_COOKIE;
    t->stack_pointer	= t->stack_high;
    bzero(t->fpu_state, FPU_STATE_SIZE);
    strcpy(t->option.name, options->name);

#ifdef SCOUT_DEBUG
    mark_blocked(t);
#endif

    return t;
} /* create */


/*
 * Retire the currently executing thread and invoke the scheduler.
 * All memory associated with the current thread is freed.
 */
static void
retire(void)
{
    dead_man = scout_thread_self;	/* let next thread deallocate space */
    scout_thread_self = 0;		/* avoid state saving */
    sched();
} /* retire */


/*
 * Invoke the initial function of thread scout_thread_self.  If it ever
 * returns, we simply retire the thread.
 */
static void
start(void)
{
    ScoutTrace(scout_thread, TR_EVENTS,
	       (__FILE__": starting thread `%s' at %p free stack %d bytes",
		scout_thread_get_name(scout_thread_self),
		scout_thread_self, scout_thread_free_stack()));

    (*scout_thread_self->fun)(scout_thread_self->option.arg);

    ScoutTrace(scout_thread, TR_EVENTS,
	       (__FILE__": thread `%s' (%#p) terminating.",
		scout_thread_get_name(scout_thread_self), scout_thread_self));
    retire();
} /* start */


void
scout_thread_init(Scout_Thread_Func user_init, void *user_arg)
{
    int prio;
    Scout_Thread_Options to;
    Scout_Thread *t;

    for (prio = 0; prio < NPRIORITIES; ++prio) {
	scout_queue_init(&ready_q[prio]);
    } /* for */
    scout_fifo_init(&async_q);
#ifdef SCOUT_DEBUG
    bzero(blocked, sizeof(blocked));
#endif

    /* create initial thread: */
    to = def_opts;
    to.name = "initializer";
    to.arg  = user_arg;
    t = create(user_init, &to);
    if (!t) {
	scout_abort("scout_thread_init: failed to create thread `%s'",
		    to.name);
    } /* if */
    scout_thread_wakeup(t);

    /* create idle thread: */
    to = def_opts;
    to.name = "idler";
    to.arg  = 0;
    to.priority = SCOUT_THREAD_PRIORITY_MIN;
    t = create(idler, &to);
    if (!t) {
	scout_abort("scout_thread_init: failed to create thread `%s'",
		    to.name);
    } /* if */
    scout_thread_wakeup(t);

    /* terminate boot stack and get things going: */
    retire();
} /* scout_thread_init */


Scout_Thread*
scout_thread_create(Scout_Thread_Func fun, const Scout_Thread_Options *options)
{
    STACK_CHECK();
    return create(fun, options);
} /* scout_thread_create */


void
scout_thread_stop(void)
{
    ScoutTrace(scout_thread, TR_EVENTS,
	       ("scout_thread_stop: thread `%s' (%p) terminating.",
		scout_thread_get_name(scout_thread_self), scout_thread_self));
    retire();
} /* Scout_ThreadStop */


void
scout_thread_suspend(void)
{
    STACK_CHECK();
#ifdef SCOUT_DEBUG
    mark_blocked(scout_thread_self());
#endif    
    sched();
} /* scout_thread_suspend */


void
scout_thread_suspend_with_continuation(void)
{
    STACK_CHECK();
#ifdef SCOUT_DEBUG
    mark_blocked(scout_thread_self());
#endif    
    scout_thread_self->running = FALSE;
    scout_thread_self->stack_pointer = scout_thread_self->stack_high;
    scout_thread_self = 0;
    sched();
} /* scout_thread_suspend_with_continuation */


void
scout_thread_wakeup(Scout_Thread *t)
{
#ifdef SCOUT_DEBUG
    mark_unblocked(t);
#endif    
    scout_assert(t->option.priority <= SCOUT_THREAD_PRIORITY_MIN);
    scout_queue_append(&ready_q[t->option.priority], t);
    if (t->option.priority < high_prio) {
	high_prio = t->option.priority;
    } /* if */
} /* scout_thread_wakeup */


void
scout_thread_async_wakeup(Scout_Thread *t)
{
    /* do not update high_prio here, we don't want to lock access to it! */
    scout_fifo_append(&async_q, t);
} /* scout_thread_async_wakeup */


void
scout_thread_yield(void)
{
    STACK_CHECK();
    scout_queue_append(&ready_q[scout_thread_self->option.priority],
		       scout_thread_self);
    sched();
} /* scout_thread_yield */


int
scout_thread_free_stack(void)
{
    char sp;

    return (&sp - (char*) &scout_thread_self->magic) - MIN_STACK_SIZE;
} /* scout_thread_free_stack */

			/*** end of scout_thread.c ***/
