/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: flamingo.c,v $
 *
 * HISTORY
 * $Log: flamingo.c,v $
 * Revision 1.10  1994/06/16  15:32:58  davidm
 * Cleaned up interrupt handling (changes to scout_device_set_interrupt(),
 * alpha_interrupt_handler(); handle_tc_error() and cb9() are new).
 *
 * Revision 1.9  1994/06/11  00:02:44  menze
 * Added copyright notice
 *
 * Revision 1.8  1994/05/29  02:25:27  druschel
 * Minor cleanup
 *
 * Revision 1.7  1994/05/25  00:59:05  druschel
 * Added support for TurboChannel option interrupts.
 *
 * Revision 1.6  1994/05/14  23:40:24  davidm
 * Minor cleanup.
 *
 * Revision 1.5  1994/05/07  18:10:59  davidm
 * Cleaned up interrupt management somewhat.
 *
 * Revision 1.4  1994/05/03  23:26:46  davidm
 * Various fixes.
 *
 * Revision 1.3  1994/04/21  07:10:23  davidm
 * (alpha_interrupt_handler): scout_age now incremented in alpha_forward.s.
 *
 * Revision 1.2  1994/04/20  00:49:30  davidm
 * Added RCS header.
 *
 */
/*
 * DEC3000 (Flamingo/Sandpiper) specific support.
 *
 * This is based on information published in DEC 3000
 * 300/400/500/600/800 Models, System Programmer's Manual, DEC Part #:
 * EK-D3SYS-PM.A01.
 */
#include <sys/types.h>
#include <machine/rpb.h>
#include "scout.h"
#include "scout_device.h"
#include "scout_time.h"
#include "scout_util.h"
#include "alpha_misc.h"
#include "alpha_pal.h"
#include "flamingo.h"
#include "turbo_channel.h"

volatile int	scout_expect_memory_access_fault = 0;
 
#define INTR_OPT0	 0
#define INTR_OPT1	 1
#define INTR_OPT2	 2
#define INTR_OPT3	 3
#define INTR_OPT4	 4
#define INTR_OPT5	 5
#define INTR_SCSI	 6
#define INTR_IOCTL	 7
#define INTR_CXTURBO	 8
#define	INTR_LANCE	 9
#define INTR_SCC0	10
#define INTR_SCC1	11
#define INTR_ISDN	12

#define INTR_NUM	13

/*
 * Interrupt dispatch table:
 */
static struct {
    void (*handler)(void*);
    void *arg;
} iswtch[INTR_NUM];

/*
 * Count number of bits set in the least significant 6 bits.
 * Based on the HAKMEM algorithm.  Assumes that at least one
 * bit in "n" is set.
 */
static inline long
cb9(u_long n)
{
    u_long t;

    t = n - ((n >> 1) & 0333) - ((n >> 2) & 0111);
    return (((t + (t >> 3)) & 07) + (t >> 6)) & 017;
} /* cb9 */


static void
handle_tc_error(u_int tcir)
{
    enum {
	SEO = (1<<0), DBF = (1<<1), X2K = (1<<2), TCR = (1<<3), TPE = (1<<4),
	TER = (1<<5), SBE = (1<<6), DBE = (1<<7),  TO = (1<<8),  TL = (1<<9),
	IA = (1<<10), NV = (1<<11), PE = (1<<12)
    };

    tcir >>= 19;
    if (tcir & SEO) {
	printf("flamingo: TURBO channel error: second error occured!\n");
    } /* if */
    if (tcir & DBF) {
	printf("flamingo: TURBO channel error: DMA buffer over/under flow!\n");
    } /* if */
    if (tcir & X2K) {
	printf("flamingo: TURBO channel error: DMA cross 2K boundary!\n");
    } /* if */
    if (tcir & TCR) {
	printf("flamingo: TURBO channel error: reset in progress!\n");
    } /* if */
    if (tcir & TPE) {
	printf("flamingo: TURBO channel error: parity error!\n");
    } /* if */
    if (tcir & TER) {
	printf("flamingo: TURBO channel error: DMA tag error!\n");
    } /* if */
    if (tcir & SBE) {
	printf("flamingo: TURBO channel error: "
	       "single bit error on I/O write or DMA read!\n");
    } /* if */
    if (tcir & DBE) {
	printf("flamingo: TURBO channel error: "
	       "double bit error on I/O write or DMA read!\n");
    } /* if */
    if (tcir & TO) {
	printf("flamingo: TURBO channel error: timeout on I/O request!\n");
    } /* if */
    if (tcir & TL) {
	printf("flamingo: TURBO channel error: DMA block too long!\n");
    } /* if */
    if (tcir & IA) {
	printf("flamingo: TURBO channel error: invalid I/O address!\n");
    } /* if */
    if (tcir & NV) {
	printf("flamingo: TURBO channel error: "
	       "scatter/gather invalid on DMA!\n");
    } /* if */
    if (tcir & PE) {
	printf("flamingo: TURBO channel error: "
	       "scatter/gather parity error!\n");
    } /* if */
} /* handle_tc_error */


/*
 * Enable or disable interrupts generated by device "di."  Interrupts
 * are enabled if "on" is true, otherwise they are disabled.  If
 * "hdlr" is not 0, set the interrupt handler to "hdlr" and its
 * argument to "arg".
 *
 * This is *NOT* an efficient way to achieve mutual exclusion.  
 * Use either atomic sequences or scout_swap_ipl() instead!
 */
void
scout_device_set_interrupt(const Scout_Device_Info *di, int on, 
	                   void (*hdlr)(void*), void *arg)
{
    Scout_IPL old_ipl;
    u_long base;
    int i;

    old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);

    base = K0SEG_TO_P((u_long)di->base_addr);
    if (base < TC_OPT0_ADDR || base > TC_OPT5_ADDR) {
	volatile u_int *const simr = (volatile u_int*) P_TO_K0SEG(SIMR);
	u_int m;

	/* junk I/O device: */
	switch (base) {
	  case SPARSE(IOCTL_LANCE_ADDR):
	    i = INTR_LANCE; m = SIR_INTR_LANCE; break;
	  case SPARSE(IOCTL_SCC0_ADDR):
	    i = INTR_SCC0; m = SIR_INTR_SCC0; break;
	  case SPARSE(IOCTL_SCC1_ADDR):
	    i = INTR_SCC1; m = SIR_INTR_SCC1; break;
	  case SPARSE(IOCTL_ISDN_ADDR):
	    i = INTR_ISDN; m = SIR_INTR_ISDN; break;
	  default: i = -1; m = 0; break;
	} /* switch */

	if (i >= 0) {
	    if (hdlr) {
		iswtch[i].handler = hdlr;
		iswtch[i].arg     = arg;
	    } /* if */
	    if (on) {
		*simr |=  m;
	    } else {
		*simr &= ~m;
	    } /* if */
	} else {
	    printf("scout_device_set_interrupt: "
		   "unknown junk I/O device at %p\n", di->base_addr);
	} /* if */
    } else {
	volatile u_int *const tcimr_rd =
	  (volatile u_int*) P_TO_K0SEG(TCIMR_READ);
	volatile u_int *const tcimr_wr =
	  (volatile u_int*) P_TO_K0SEG(TCIMR_WRITE);

	switch (base) {
	  case TC_OPT0_ADDR: i = INTR_OPT0; break;
	  case TC_OPT1_ADDR: i = INTR_OPT1; break;
	  case TC_OPT2_ADDR: i = INTR_OPT2; break;
	  case TC_OPT3_ADDR: i = INTR_OPT3; break;
	  case TC_OPT4_ADDR: i = INTR_OPT4; break;
	  case TC_OPT5_ADDR: i = INTR_OPT5; break;
	  default: i = -1; break;
	} /* switch */
	if (i >= 0) {
	    if (hdlr) {
		iswtch[i].handler = hdlr;
		iswtch[i].arg     = arg;
	    } /* if */
	    if (on) {
		*tcimr_wr = *tcimr_rd & ~(1<<i);
	    } else {
		*tcimr_wr = *tcimr_rd |  (1<<i);
	    } /* if */
	} else {
	    printf("scout_device_set_interrupt: "
		   "unknown TURBOchannel device at %p\n", di->base_addr);
	} /* if */
    } /* if */
    scout_swap_ipl(old_ipl);
} /* scout_device_set_interrupt */


void
alpha_interrupt_handler(u_long type, u_long vector, long *logout_area)
{
    Scout_Thread *t;
    u_long sir, tcir;

    switch (type) {
      case 0:
	printf("got interprocessor interrupt!\n"); break;
      case 1:
	t = (Scout_Thread*) scout_clock_thread;
	if (t) {
	    scout_clock_thread = 0;
	    scout_thread_async_wakeup(t);
	} /* if */
	break;
      case 2:
	if (scout_expect_memory_access_fault) {
	    mtpr_mces(0x19);
	    scout_expect_memory_access_fault = 0;
	} else {
	    scout_panic("flamingo_interrupt: got unexpected machine check!");
	} /* if */
	break;
      case 3:
	sir = *(u_int*)P_TO_K0SEG(SIR) & *(u_int*)P_TO_K0SEG(SIMR);
	if (sir) {
	    if (sir & SIR_INTR_LANCE) {
		(*iswtch[INTR_LANCE].handler)(iswtch[INTR_LANCE].arg);
	    } else if (sir & SIR_INTR_SCC0) {
		(*iswtch[INTR_SCC0].handler)(iswtch[INTR_SCC0].arg);
	    } else if (sir & SIR_INTR_SCC1) {
		(*iswtch[INTR_SCC1].handler)(iswtch[INTR_SCC1].arg);
	    } else if (sir & SIR_INTR_ISDN) {
		(*iswtch[INTR_ISDN].handler)(iswtch[INTR_ISDN].arg);
	    } /* if */
	} else {
	    /*
	     * Read and clear TC interrupt register and mask it with
	     * (complement of) TC interrupt mask register to get pending
	     * interrupts.  As we clear *all* interrupt requests, we
	     * better process all of them at once (this is an unlikely
	     * event though, and we optimize for the common case of only
	     * one interrupt being pending).
	     */
	    tcir = *(u_int*)P_TO_K0SEG(TCIR_CLEAR)
	      & ~(*(u_int*)P_TO_K0SEG(TCIMR_READ));
#ifdef SCOUT_DEBUG
	    if (!tcir) {
	        printf("got a spurious interrupt, sir=%x, tcir=%x\n",
		       (u_int) sir, (u_int) tcir);
	    } /* if */
#endif
	    if (tcir & ~0777) {
		handle_tc_error(tcir);
		tcir &= 0777;
	    } /* if */
	    while (tcir) {
		u_long lsb, i;

		lsb = tcir & ~(tcir - 1);
		tcir &= ~lsb;
		i = cb9(lsb - 1);
		scout_assert(i < 9);
		(*iswtch[i].handler)(iswtch[i].arg);
	    } /* while */
	} /* if */
	break;
      case 4:
	printf("got performance counter interrupt\n"); break;
      default:
	printf("got unknown I/O device interrupt (type=%ld,vector=%ld)\n",
	       type, vector);
    } /* switch */
} /* alpha_interrupt_handler */


void
alpha_instruction_fault(u_long type, Alpha_Saved_State *ss)
{
    switch (type) {
      case 0:
	kgdb_handle_exception(0, ss);
	break;
      case 1:
	printf("hit bugcheck!\n");
	kgdb_handle_exception(1, ss);
	break;
      case 2:
	printf("hit gentrap!\n");
 	kgdb_handle_exception(2, ss);
	break;
      case 3:
	printf("hit FEN fault!\n"); break;
      case 4:
	printf("hit opDEC!\n");
	kgdb_handle_exception(4, ss);
	break;
    } /* switch */
} /* alpha_instruction_fault */


void
alpha_mm_fault(u_long vaddr, u_long mmcsr, long access, Alpha_Saved_State *ss)
{
    if (scout_expect_memory_access_fault) {
	scout_expect_memory_access_fault = 0;
	ss->pc += 4;	/* just skip faulting instruction */
    } else {
	printf("alpha_mm_fault(vaddr=%lx,mmcsr=%ld,access=%ld,ss=%p,pc=%p)\n",
	       vaddr, mmcsr, access, ss, ss->pc);
	scout_panic("unexpected access fault");
    } /* if */
} /* alpha_mm_fault */


void
alpha_unaligned_fault(u_long vaddr, int opcode, int reg, Alpha_Saved_State *ss)
{
    printf("alpha_unaligned_fault(vaddr=%#lx,op=%#x,reg=%d,ss=%p,pc=%p)\n",
	   vaddr, opcode, reg, ss, ss->pc);
    scout_panic("unaligned access fault");
} /* alpha_unaligned_fault */


int
read_int(volatile int *pa)
{
    return *pa;
} /* read_int */


/*
 * Test whether a given K0SEG address is read accessible.  Return TRUE
 * if yes, FALSE if not.  The assumption is that a machine check will
 * be generated if the address cannot be read.  This is mostly useful
 * for checking the presence of TURBOchannel options.
 */
int
probe_read(void *pa)
{
    scout_expect_memory_access_fault = TRUE;
    read_int(pa);
    if (!scout_expect_memory_access_fault) {
	return FALSE;
    } /* if */
    scout_expect_memory_access_fault = FALSE;
    return TRUE;
} /* probe_read */


char*
lance_config_str(void)
{
    u_int ssr = *(volatile u_int*)P_TO_K0SEG(SSR);

    if (ssr & SSR_LANCE_10BASET) {
	if (ssr & SSR_LANCE_LOOPBACK_OFF) {
	    return "10Base-T";
	} else {
	    return "10Base-T local loopback";
	} /* if */
    } else {
	if (ssr & SSR_LANCE_LOOPBACK_OFF) {
	    return "Thickwire";
	} else {
	    return "Thickwire local loopback";
	} /* if */
    } /* if */
} /* lance_config_str */


void
lance_enable_dma(void *shared_memory_ptr)
{
    scout_assert(((u_long) shared_memory_ptr & 0xffff) == 0);
    /*
     * Top 5 bits of 34 bit physical address go to bits 0:4,
     * middle 12 bits go to bits 20:31 and remaining bits
     * are generated by LANCE:
     */
    *(volatile u_int*)P_TO_K0SEG(LANCE_DMA_PTR) = 
      (((u_long) shared_memory_ptr >> 29) & 0x1f)
	| ((u_long) shared_memory_ptr >> 17) << 20;
    *(volatile u_int*)P_TO_K0SEG(SSR) |= SSR_LANCE_DMA_ENABLE;
} /* lance_enable_dma */


static void
default_handler(void *arg)
{
    int i = (long)arg;

    printf("flamingo: got interrupt for device %d without a handler!\n", i);
} /* default_handler */


void
flamingo_init(void)
{
    volatile u_int *const simr = (volatile u_int*) P_TO_K0SEG(SIMR);
    volatile u_int *const tcimr_wr = (volatile u_int*) P_TO_K0SEG(TCIMR_WRITE);
    Scout_Device_Info di;
    int i;

    for (i = 0; i < INTR_NUM; ++i) {
	iswtch[i].handler = default_handler;
	iswtch[i].arg     = (void*) (long) i;
    } /* for */

    *simr = 0;		/* disable junk I/O interrupts */
    *tcimr_wr = 077;	/* disable TC interrupts except for errors */

    /* reset LANCE: */
    *(volatile u_int*)P_TO_K0SEG(SSR) &= ~SSR_LANCE_RESET_OFF;

    /* configure TURBOchannel (all I/O interrupts are at level 4): */
    tc_config((void*) P_TO_K0SEG(TC_BASE_ADDR), TC_SLOT_SIZE, 6, 4);

    /* add Flamingo/Sandpiper I/O board devices: */

    di.name = "scsi";	/* NCR 53C94 SCSI chip(s) */
    di.base_addr = (void*) P_TO_K0SEG(SPARSE(IOCTL_SCSI_ADDR));
    di.rom_offset = 0;
    di.ipl = 4;
    scout_device_add(&di);

    di.name = "ln";		/* AMD 7990 LAN Controller for Ethernet */
    di.base_addr  = (void*) P_TO_K0SEG(SPARSE(IOCTL_LANCE_ADDR));
    di.rom_offset = -0x80000L;
    di.ipl = 4;
    scout_device_add(&di);

    di.name = "scc";	/* Zilog 8530 Serial Controller Chip */
    di.base_addr = (void*) P_TO_K0SEG(SPARSE(IOCTL_SCC0_ADDR));
    di.rom_offset = 0;
    di.ipl = 4;
    scout_device_add(&di);

    di.name = "scc";	/* Zilog 8530 Serial Controller Chip */
    di.base_addr = (void*) P_TO_K0SEG(SPARSE(IOCTL_SCC1_ADDR));
    di.rom_offset = 0;
    di.ipl = 4;
    scout_device_add(&di);

    di.name = "rtc";	/* DS 1287A real-time clock chip */
    di.base_addr = (void*) P_TO_K0SEG(SPARSE(IOCTL_RTC_ADDR));
    di.rom_offset = 0;
    di.ipl = 4;
    scout_device_add(&di);

    di.name = "isdn";	/* AMD 7930 ISDN chip */
    di.base_addr = (void*) P_TO_K0SEG(SPARSE(IOCTL_ISDN_ADDR));
    di.rom_offset = 0;
    di.ipl = 4;
    scout_device_add(&di);

    /* take LANCE off of reset: */
    *(volatile u_int*)P_TO_K0SEG(SSR) |= SSR_LANCE_RESET_OFF;

#ifdef SCOUT_KGDB
    scc_enable_interrupts();	/* XXX fix */
#endif
} /* flamingo_init */

			/*** end of flamingo.c ***/
