/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scc.c,v $
 *
 * HISTORY
 * $Log: scc.c,v $
 * Revision 1.5  1994/06/16  15:37:40  davidm
 * Changed interrupt handling according to flamingo.c changes.
 *
 * Revision 1.4  1994/06/11  00:02:47  menze
 * Added copyright notice
 *
 * Revision 1.3  1994/05/07  18:12:03  davidm
 * (scc_attach): interrupts are now enabled in flamingo.c.
 *
 * Revision 1.2  1994/05/03  23:26:18  davidm
 * (scc0_interrupt): added.
 *
 * It is now possible to interrupt a running kernel from an attached
 * gdb.
 *
 * Revision 1.1  1994/05/03  17:42:08  davidm
 * Initial revision
 *
 */
/*
 * Use as few functions as possible!  If a breakpoint is set in
 * any of those functions, the debugger will crash because it
 * will cause infinite recursion!!
 */
#include <machine/rpb.h>
#include <sys/types.h>
#include "kgdb.h"
#include "scout_device.h"
#include "scout_sync_arch.h"
#include "alpha_misc.h"
#include "flamingo.h"

#define CHAN_B	0
#define CHAN_A	1

#define SCC_TX_EMPTY		0x04
#define SCC_RX_READY		0x01

#define	RESET_ERROR		0x30
#define	RESET_HIGHEST_IUS	0x38	/* channel A only */

#define CTS		0x20
#define DCD		0x08

#define	DTR		0x80	/* drive DTR pin */
#define	TX8		0x60
#define	TX_ENABLE	0x08
#define	RTS		0x02	/* drive RTS pin */

#define	FRAME_ERR	0x40	/* bad frame */
#define	RX_OVERRUN	0x20	/* rcv fifo overflow */
#define	PARITY_ERR	0x10	/* incorrect parity in data */

#define	RXI_SPECIAL_O	0x18	/* on special only */
#define	RXI_ALL_CHAR	0x10	/* on each char, or special */
#define	RXI_FIRST_CHAR	0x08	/* on first char, or special */
#define	RXI_DISABLE	0x00	/* never on recv */

#define	MASTER_IE	0x08


typedef volatile struct {
    struct {
	u_long command;		/* register select */
	u_long data;		/* rx/tx buffer */
    } channel[2];
} SCC_Register_Map;


static SCC_Register_Map *scc_chip[2] = {
    (SCC_Register_Map*) P_TO_K0SEG(SPARSE(IOCTL_SCC0_ADDR)),
    (SCC_Register_Map*) P_TO_K0SEG(SPARSE(IOCTL_SCC1_ADDR))
};

#define scc_read_reg(scc,chan,reg) \
  ((scc)->channel[(chan)].command = (reg) << 8, \
   ((scc)->channel[(chan)].command >> 8))

#define scc_write_reg(scc,chan,reg,val) \
{ \
   (scc)->channel[(chan)].command = (reg) << 8; \
   asm volatile ("mb" ::: "memory"); \
   ((scc)->channel[(chan)].command = ((val) & 0xff) << 8); \
   asm volatile ("mb" ::: "memory"); \
}


/*
 * Invariant: rx_rdp == rx_wrp => buffer empty; implies we can store
 * sizeof(rx_buf)-1 characters only.  Who cares?
 */
static char rx_buf[1024];
static int rx_rdp = 0;
static volatile int rx_wrp = 0;


static void
udelay(long usec)
{
    u_int start, delta;
    static int cycles_per_usec = 0;

    if (!cycles_per_usec) {
	cycles_per_usec = rpb->rpb_counter / 1000000L;
    } /* if */
    delta = usec * cycles_per_usec;
    start = read_itimer();
    while (read_itimer() - start < delta) {
	/* skip */
    } /* while */
} /* udelay */


void
scc_attach(void)
{
    SCC_Register_Map *scc;

    scc = scc_chip[0];
    scc_write_reg(scc, CHAN_B, 12, 4);		/* set baudrate to 38400 */
    scc_write_reg(scc, CHAN_B, 13, 4 >> 8);
    scc_write_reg(scc, CHAN_B,  1, RXI_ALL_CHAR);	/* enable rx intrs */
    scc_write_reg(scc, CHAN_B,  9, MASTER_IE);	/* enable master i-enable */
    scc_write_reg(scc, CHAN_A,  0, RESET_HIGHEST_IUS);
} /* scc_attach */


void
scc_putc(int ch)
{
    SCC_Register_Map *scc = scc_chip[0];
    u_char value;

    /* wait until TX buffer empty: */
    do {
	value = scc_read_reg(scc, CHAN_B, 0);
	udelay(100);
    } while ((value & (SCC_TX_EMPTY|CTS|DCD)) != (SCC_TX_EMPTY|CTS|DCD));
    scc->channel[CHAN_B].data = (u_char) ch << 8;
} /* scc_putc */


int
scc_getc(void)
{
    Scout_IPL old_ipl;
    int ch;

    old_ipl = scout_swap_ipl(SCOUT_MAX_IPL);
    scout_swap_ipl(old_ipl);

    if ((rx_rdp != rx_wrp) || (old_ipl < 4)) {
	/* read character interrupt driven: */
	while (rx_rdp == rx_wrp) {
	    /* skip */
	} /* while */
	ch = rx_buf[rx_rdp];
	rx_rdp = (rx_rdp + 1) % sizeof(rx_buf);
    } else {
	SCC_Register_Map *scc = scc_chip[0];
	u_long value;

	/* read character in polling mode: */
      again:
	value = scc->channel[CHAN_B].command >> 8;
	if (!(value & SCC_RX_READY)) {
	    /* enable transmission: */
	    scc_write_reg(scc, CHAN_B, 5, DTR | TX8 | TX_ENABLE | RTS);
	    do {
		value = scc->channel[CHAN_B].command >> 8;
	    } while (!(value & SCC_RX_READY));
	    /* disable transmission: */
	    scc_write_reg(scc, CHAN_B, 5, TX8 | TX_ENABLE);
	} /* if */

	value = scc_read_reg(scc, CHAN_B, 1);
	ch = (scc->channel[CHAN_B].data >> 8) & 0xff;
	if (value & (PARITY_ERR | RX_OVERRUN | FRAME_ERR)) {
	    scc_write_reg(scc, CHAN_B, 0, RESET_ERROR);
	    scc_write_reg(scc, CHAN_A, 0, RESET_HIGHEST_IUS);
	    goto again;
	} /* if */
	scc_write_reg(scc, CHAN_A, 0, RESET_HIGHEST_IUS);
    } /* if */
    return ch;
} /* scc_getc */


static void
interrupt(void *arg)
{
    SCC_Register_Map *scc = scc_chip[0];
    u_long value;
    int ch;

    value = scc->channel[CHAN_B].command >> 8;
    if (value & SCC_RX_READY) {
	value = scc_read_reg(scc, CHAN_B, 1);
	ch = (scc->channel[CHAN_B].data >> 8) & 0xff;
	if (value & (PARITY_ERR | RX_OVERRUN | FRAME_ERR)) {
	    /* error during receive---reset error: */
	    scc_write_reg(scc, CHAN_B, 0, RESET_ERROR);
	} else {
	    /* received character OK: */
	    if (ch == KGDB_BREAK_CHARACTER) {
		scc_write_reg(scc, CHAN_A, 0, RESET_HIGHEST_IUS);
		printf("kgdb interrupt...\n");
		kgdb_breakpoint();
		return;
	    } else if ((rx_wrp + 1) % sizeof(rx_buf) != rx_rdp) {
		rx_buf[rx_wrp] = ch;
		rx_wrp = (rx_wrp + 1) % sizeof(rx_buf);
	    } /* if */
	} /* if */
	/*
	 * Reset "highest interrupt under service" to ACK handling of
	 * this interrupt:
	 */
	scc_write_reg(scc, CHAN_A, 0, RESET_HIGHEST_IUS);
    } /* if */
} /* interrupt */


/*
 * Hack to enable interrupts *after* device manager is initialized.
 */
void
scc_enable_interrupts(void)
{
    const Scout_Device_Info *di;

    di = scout_device_info("scc0");
    if (di) {
	scout_device_set_interrupt(di, 1, interrupt, 0);
    } else {
	xError("scc_init: couldn't find scc0---gdb can't interrupt");
    } /* if */
} /* scc_enable_interrupts */

			/*** end of scc.c ***/
